var request = require('request');
var querystring = require('querystring');


/**
 * Empty function for use when a callback isn't supplied.
**/
function noop() {}


/**
 * Takes a given node of an object and formats it accordingly
 *
 * @param {Mixed}      node        The given node of an object.
 * @private
**/
function format(node) {
  if (Array.isArray(node)) {
    return formatObject(node).join(',');
  } else if (typeof node == 'object') {
    return JSON.stringify(node);
  } else {
    return node;
  }
}


/**
 * Takes a given object and formats it according to format()
 *
 * @param {Object}      obj        The object to format.
 * @return {Object} The formatted object.
 * @private
**/
function formatObject(obj) {
  var keys = Object.keys(obj), result = Array.isArray(obj) ? [] : {};
  for (var i = 0, l = keys.length; i < l; i++) {
    result[keys[i]] = format(obj[keys[i]]);
  }
  return result;
}



/**
 * Creates a new Facebook Graph Client with the optional access_token.
 *
 * @param  {String}  access_token  Optional user access_token.
 * @constructor
**/
function Client(access_token) {
  this.access_token = access_token || null;
  this.host = 'graph.facebook.com';
}


/**
 * Publicly export Client.
**/
exports.Client = Client;


/**
 * Publicly export Client.
 *
 * @see Client
 * @return {Client} The new Client instance.
**/
exports.createClient = function(access_token) {
  return new Client(access_token);
};


/**
 * Sends a POST request for an Object on the Graph
 *
 * @param {String}    id          The facebook users' id (or 'me').
 * @param {Object}    args        Any extra arguments for the querystring.
 *                                Eg, since, until, access_token, etc.
 * @param {Function}  cb          Callback function.
**/
Client.prototype.postObject = function(id, args, cb) {
  this.request('post', id, args, cb);
};


/**
 * Sends a DELETE request for an Object on the Graph
 *
 * @param {String}    id          The facebook users' id (or 'me').
 * @param {Object}    args        Any extra arguments for the querystring.
 *                                Eg, since, until, access_token, etc.
 * @param {Function}  cb          Callback function.
**/
Client.prototype.deleteObject = function(id, args, cb) {
  this.request('delete', id, args, cb);
};


/**
 * Sends a GET request for an Object on the Graph
 *
 * @param {String}    id          The facebook users' id (or 'me').
 * @param {Object}    args        Any extra arguments for the querystring.
 *                                Eg, since, until, access_token, etc.
 * @param {Function}  cb          Callback function.
**/
Client.prototype.getObject = function(id, args, cb) {
  this.request('get', id, args, cb);
};


/**
 * Sends a GET request for multiple Objects on the Graph
 *
 * @param {String[]}  ids         The facebook user ids.
 * @param {Object}    args        Any extra arguments for the querystring.
 *                                Eg, since, until, access_token, etc.
 * @param {Function}  cb          Callback function.
**/
Client.prototype.getObjects = function(ids, args, cb) {
  if (!cb && typeof args == 'function') {
    cb = args;
    args = {};
  } else if (!cb && typeof args != 'function') {
    cb = noop;
  }

  args.ids = ids;
  this.request('get', '', args, cb);
};


/**
 * Sends a GET request for a Connection of an Object on the Graph
 *
 * @param {String}    id          The facebook users' id (or 'me').
 * @param {String}    connection  The graph connection, eg 'feed'.
 * @param {Object}    args        Any extra arguments for the querystring.
 *                                Eg, since, until, access_token, etc.
 * @param {Function}  cb          Callback function.
**/
Client.prototype.getConnection = function(id, connection, args, cb) {
  this.request('get', id + '/' + connection, args, cb);
};


/**
 * Sends a POST request to a Connection of an Object on the Graph
 *
 * @param {String}    id          The facebook users' id (or 'me').
 * @param {String}    connection  The graph connection, eg 'feed'.
 * @param {Object}    args        Any extra arguments for the querystring.
 *                                Eg, since, until, access_token, etc.
 * @param {Function}  cb          Callback function.
**/

Client.prototype.postConnection = function(id, connection, args, cb) {
  this.request('post', id + '/' + connection, args, cb);
};


/**
 * Sends a new request to facebook with the given method, path,
 * args, and triggers the cb on completion.
 *
 * @param {String}      method  The HTTP method for the request.
 * @param {String|Null} path    The path for the request, null for '/'.
 * @param {Object}      args    The optional arguments for the request.
 *        If the request is a POST request, args is the body of the request.
 * @param {Function}    cb      The function to call on completion.
 *        The function's signature is f(error, response, data).
**/
Client.prototype.request = function(method, path, args, cb) {
  var options = {
    uri: 'https://' + this.host + '/' + path,
    method: method.toUpperCase()
  };

  if (!cb && typeof args == 'function') {
    cb = args;
    args = {};
  } else if (!cb && typeof args != 'function') {
    cb = noop;
  }

  if (!args.access_token && this.access_token) {
    args.access_token = this.access_token;
  }

  /*
    For requests like POST where you have args like privacy, it requires
    the value to be a json string of an object, this is more easily written
    as a JS Object, and then automatically converted.

    eg.
      {
        message: 'test',
        privacy: {value: 'ALL_FRIENDS'},
        actions: {"name": "View on Zombo", "link": "http://www.zombo.com"}
      }

    would need to become:
      {
        message: 'test',
        privacy: '{"value":"ALL_FRIENDS"}',
        actions: '{"name":"View on Zombo","link":"http://www.zombo.com"}'
      }
  */

  args = formatObject(args);

  if (options.method == 'POST') {
    options.body = querystring.stringify(args);
  } else {
    options.uri += '?' + querystring.stringify(args);
  }

  this._request(options, cb);
};


/**
 * Actually perform the request, this is stubbed out for testing purposes
 *
 * @param {Object}      options     The request options.
 * @param {Function}    cb          The request callback.
**/
Client.prototype._request = function(options, cb) {
  request(options, function(err, res, data) {
    if (err) { // HTTP Error
      cb(err, res, data);
    } else {
      // JSON can often throw while parsing,
      // so let's not kill our node app due to that.
      try {
        data = JSON.parse(data);

        if (data.error) {
          var e = new Error(data.error.message);
          e.name = data.error.type;

          cb(e, res, data);
        } else {
          cb(null, res, data);
        }
      } catch (e) {
        cb(e, res, data);
      }
    }
  });
};
